/*****************************************************************
   We need to find out the list of files in the local directory. 
   This list will later be stored on the remote server. 
*****************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <zlib.h>
#include "eprintf.h"
#include "files.h"
#include "ptrbuf.h"



static void list_files(char *dirname, ptrbuf_t *files, ptrbuf_t *dirs);
static void fileset_add
   (ptrbuf_t *dst, uint64_t mtime, uint64_t size, char *name);
static void combine_fn(char **Rdst, int *Rsiz, char *dir, char *file);
static size_t serialize_entry(unsigned char *tgt, file_t *f);
static size_t write_int(unsigned char *tgt, uint64_t V, int sz);
static uint64_t read_int(unsigned char *src, int sz);
static size_t find_str_end(unsigned char *str, unsigned char *end);

static void list_files(char *dirname, ptrbuf_t *files, ptrbuf_t *dirs)
{
  DIR *dir;
  struct dirent *de;
  struct stat st;
  static char *fname;
  static int fname_size;

  dir= opendir(dirname? dirname: ".");
  if (!dir)
  {
    eprintf("opendir(%s): %s\n", dirname? dirname : ".", strerror(errno));
    return ;
  }

  while((de=readdir(dir)))
  {
    if (!strcmp(de->d_name,"dsync.cfg") || !strcmp(de->d_name,".") ||
        !strcmp(de->d_name,"..")) continue;

    combine_fn(&fname, &fname_size, dirname, de->d_name);

    if (stat(fname, &st))
    {
      eprintf("stat %s: %s\n", fname, strerror(errno));
      continue;
    }
    if (S_ISREG(st.st_mode))
      fileset_add(files, st.st_mtime, st.st_size, fname);
    else if (S_ISDIR(st.st_mode))
      fileset_add(dirs, st.st_mtime, 0, fname);
  }
  closedir(dir);
}


static void fileset_add
   (ptrbuf_t *dst, uint64_t mtime, uint64_t size, char *name)
{
  ptrbuf_put(dst, file_new(mtime, size, name));
}


static void combine_fn(char **Rdst, int *Rsiz, char *dir, char *file)
{
  int sz;
  if (dir)
     sz= strlen(dir) + 1 + strlen(file) + 1;
  else
     sz= strlen(file) + 1;
  if (sz>*Rsiz)
  {
    if (*Rdst) free(*Rdst);
    *Rdst= malloc(sz);
    *Rsiz= sz;
  }
  if (dir)
    sprintf(*Rdst,"%s/%s",dir,file);
  else
    strcpy(*Rdst, file);
}

void list_all_files(file_t ***Rfiles, size_t *Rnfiles,
                           file_t ***Rdirs, size_t *Rndirs)
{
  ptrbuf_t *pfile, *pdir, *ptempdir;
  file_t *D;

  pfile= ptrbuf_new(100);
  ptempdir= ptrbuf_new(100);
  pdir= ptrbuf_new(100);

  list_files(NULL, pfile, ptempdir);

  while((D=ptrbuf_get(ptempdir)))
  {
    ptrbuf_put(pdir, D);
    list_files(D->name, pfile, ptempdir);
  }

  ptrbuf_free(ptempdir, NULL, NULL);
  ptrbuf_free(pfile, (void***) Rfiles, Rnfiles);
  ptrbuf_free(pdir, (void***) Rdirs, Rndirs);
  files_sort(*Rfiles,*Rnfiles);
  files_sort(*Rdirs,*Rndirs);
}

void serialize_file_list(file_t **files, size_t nfiles, 
                         unsigned char **Rbuffer, size_t *Rlen)
{
  size_t raw_size, i,offset,comp_len;
  unsigned char *raw_buffer, *comp_buffer;
  
  raw_size= 0;

  for(i=0;i<nfiles;i++)
    raw_size+= strlen(files[i]->name) + 1 + 8;

     // 8 bytes for headers and 1 byte signature. 'R'= raw 'Z'=zipped
  raw_buffer= malloc(raw_size + 9);
  offset= 9;
  for(i=0;i<nfiles;i++)
    offset+= serialize_entry(raw_buffer+offset, files[i]);

  comp_buffer= malloc(raw_size+1);
  comp_len= raw_size+1;

  if (compress2(comp_buffer, &comp_len, raw_buffer+9,
                raw_size,Z_BEST_COMPRESSION)!=Z_OK)
  {
     raw_buffer[0]= 'R';
     *Rlen= raw_size+9;
  }
  else
  {
     free(raw_buffer);
     raw_buffer= malloc(comp_len+9);
     memcpy(raw_buffer+9, comp_buffer, comp_len);
     raw_buffer[0]= 'Z';
     *Rlen= comp_len+9;
  }

  write_int(raw_buffer+1, nfiles, 4);
  write_int(raw_buffer+5, raw_size, 4);
  *Rbuffer= raw_buffer;
  free(comp_buffer);
}

static size_t serialize_entry(unsigned char *tgt, file_t *f)
{
  size_t L;
  write_int(tgt, f->mtime, 8);
  L= strlen(f->name);
  memcpy(tgt+8,f->name, L+1);
  return L+1+8;
}

static size_t write_int(unsigned char *tgt, uint64_t V, int sz)
{
  int i;
  for(i=0;i<sz;i++)
   tgt[i]= (V>>(i*8))&0xff;
  return i;
}

static uint64_t read_int(unsigned char *src, int sz)
{
  uint64_t V,P;
  int i;
  V= 0;
  for(i=0;i<sz;i++)
  {
    P= src[i];
    V|= P<<(i*8);
  }
  return V;
}


int deserialize_file_list(unsigned char *buffer, size_t len,
                           file_t ***Rfiles, size_t *Rnfiles)
{
  unsigned char *fbuffer;
  size_t raw_size, fsize,i, nfiles,offset;
  file_t **files;

  nfiles= read_int(buffer+1,4);
  raw_size= read_int(buffer+5,4);
  if (buffer[0]=='Z')
  {
    fbuffer= malloc(raw_size);
    if (uncompress(fbuffer, &fsize, buffer+9, len-9)!=Z_OK || fsize!=raw_size)
    {
      free(fbuffer);
      return 1;
    }
  }
  else if (buffer[0]=='R')
  {
    fbuffer= buffer+9;
  }
  else
  {
    return 1;  // bad magic
  }

  files= malloc(sizeof(files[0])*nfiles);
  offset= 0;
  for(i=0;i<nfiles && offset<raw_size;i++)
  {
    uint64_t tm;
    size_t L;
    tm= read_int(fbuffer+offset,8);
    L= find_str_end(fbuffer+offset+8, fbuffer+raw_size);
    if (!L) break;
    files[i]= file_new(tm, 0, (char*) fbuffer+offset+8);
    offset+= 8+L+1;
  }
  if (fbuffer!=buffer+9)
    free(fbuffer);
  *Rfiles= files;
  *Rnfiles= nfiles;
  files_sort(files,nfiles);
  return 0;
}

static size_t find_str_end(unsigned char *str, unsigned char *end)
{
  size_t L;
  for(L=0;L+str<end;L++)
    if (str[L]==0) return L;
  return 0;
}

file_t *file_new(uint64_t mtime,uint64_t size, char *name)
{
  file_t *F;
  size_t L;
  L= strlen(name);
  F= calloc(1,sizeof(*F)+L);
  F->mtime= mtime;
  F->size= size;
  memcpy(F->name, name, L+1);
  return F;
}

static int file_comp(const void *a, const void *b)
{
  const file_t *A= *(const file_t **) a;
  const file_t *B= *(const file_t **) b;
  return strcmp(A->name, B->name);
}

void files_sort(file_t **files, size_t nfiles)
{
  size_t i;
  if (nfiles)
    for(i=0;i<nfiles-1;i++)
     if (strcmp(files[i]->name, files[i+1]->name)>0)
     { 
       qsort(files, nfiles, sizeof(files[0]), file_comp);
       break;
     }
}

void files_unify(file_t **files, size_t *IRnfiles)
{
   size_t i,j;
   size_t nfiles;
   nfiles= *IRnfiles;
   if (!nfiles) return ;
   j= 0;
   for(i=1;i<nfiles;i++)
     if (strcmp(files[i]->name,files[j]->name))
     {
       j++;
       files[j]= files[i];
     } 
   *IRnfiles= j+1;
}
