#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdint.h>
#include <string.h>
#include <signal.h>
#include "eprintf.h"
#include "config.h"
#include "target.h"
#include "files.h"
#include "ptrbuf.h"
#include "remodirs.h"

int force_compare;
static int command; 
cfg_target_t *target;

void writef(void *F, char *K, size_t L)
{
  fwrite(K, 1, L, F);
}


static void parse_args(int argc, char **argv)
{
   int i;
   char *tgtname;
   if (argc==2)
   {
     command= 'U';
     tgtname= argv[1];
     goto done;
   }
   if (argc==3)
   {
     if (!strcmp(argv[1],"update")) { command= 'U'; tgtname= argv[2]; 
                                      goto done; }
     if (!strcmp(argv[1],"force")) { command= 'F'; tgtname= argv[2]; 
                                      goto done; }
     if (!strcmp(argv[1],"mark")) { command= 'M'; tgtname= argv[2]; 
                                      goto done; }
     if (!strcmp(argv[1],"list")) { command= 'L'; tgtname= argv[2]; 
                                      goto done; }
     goto usage;
   }
   goto usage;
done:
   for(i=0;i<settings.N_target;i++)
       if (!strcmp(settings.target[i].name, tgtname)) break;
   if (i==settings.N_target)
   {
     eprintf("can't find target %s in configuration file.\n", tgtname);
     goto usage;
   }
   target= settings.target+i;
   return ;
usage:
   eprintf("usage: dsync command target\n"
           "   command is one of:\n"
           "     update     Updates the modified files on target.\n"
           "     force      Uploads all files in the directory to the target.\n"
           "     mark       Marks the target as up-to-date, without uploading"
         "\n                any files other than the file list.\n"
           "     list       lists which files need to be updated on the target."
         "\n                Nothing is uploaded.\n"
           "     help       Displays a help page, no target should be given."
           "   If no command is given, it's assumed to be 'update'.\n");
   exit(1);
}

static int sign(int x) { return x<0? -1 : (x>0? 1:0); }

static void file_set_split(file_t **A, size_t nA, file_t **B, size_t nB,
                         file_t ***RaMb, size_t *RNaMb,
                         file_t ***RaIb, size_t *RNaIb,
                         file_t ***RbMa, size_t *RNbMa,
                         file_t ***RbIa, size_t *RNbIa)
{
  size_t i,j;
  ptrbuf_t *aMb, *aIb, *bMa, *bIa;

  aMb= ptrbuf_new(100);
  aIb= ptrbuf_new(100);
  bMa= ptrbuf_new(100);
  bIa= ptrbuf_new(100);

  i= 0; j= 0;
  while(i<nA && j<nB)
    switch(sign(strcmp(A[i]->name,B[j]->name)))
    {
    case 0:   
      ptrbuf_put(aIb, A[i]);
      ptrbuf_put(bIa, B[j]);
      i++;
      j++;
      break;
    case 1:
      ptrbuf_put(bMa, B[j]);
      j++;
      break;
    case -1:
      ptrbuf_put(aMb, A[i]);
      i++;
      break;
    }
  for(;i<nA;i++) ptrbuf_put(aMb, A[i]);
  for(;j<nB;j++) ptrbuf_put(bMa, B[j]);

  ptrbuf_free(aMb, (void***) RaMb, RNaMb);
  ptrbuf_free(aIb, (void***) RaIb, RNaIb);
  ptrbuf_free(bMa, (void***) RbMa, RNbMa);
  ptrbuf_free(bIa, (void***) RbIa, RNbIa);
}



static int get_remote_file_list(file_t ***Rfiles, size_t *Rnfiles)
{
   uint8_t *buffer; size_t len;
   int R;
   size_t nfiles; file_t **files;

   buffer= NULL;
   len= 0;
   *Rfiles= NULL;
   *Rnfiles= 0;

   if (read_remote_file("files.lst", &buffer, &len))
   {
     if (buffer) free(buffer);
     return 1;
   }
   R= deserialize_file_list(buffer, len, Rfiles, Rnfiles);
   free(buffer);
   nfiles= *Rnfiles;
   files= *Rfiles;
   if (R || nfiles==0) 
   {
     if (files) free(files);
     *Rfiles= NULL;
     *Rnfiles= 0;
     return 1;
   }

   // This will never actually find any duplicates, but since I'm
   // going to report removable files, I should do this anyway.
   // Just in case something miraculous happens and the file list
   // update is interrupted in the middle which leaves it with
   // two entries for the same file.
   files_unify(files, Rnfiles);
   return 0;
}

void show_help()
{
  puts("HELP");
}


void print_file_list(const char *msg,file_t **files, size_t nfiles)
{
  size_t i;
  printf("%s:",msg);
  if (nfiles==0) printf(" None");
  else for(i=0;i<nfiles;i++)
      printf("\n  %s", files[i]->name);
  printf("\n");
}

int write_remote_list(file_t **files, size_t nfiles)
{
  uint8_t *buffer; size_t len;
  int R;
  printf("!updating remote index\n");
  serialize_file_list(files,nfiles, &buffer, &len);
  R= write_remote_file("files.lst", buffer,len);
  free(buffer);
  if (R) eprintf("failed to write remote index\n"); 
  return R;
}

void do_list()
{
  size_t i;
  file_t **files, **dirs, **rfiles;
  size_t nfiles, ndirs, nrfiles;
  list_all_files(&files, &nfiles, &dirs, &ndirs);
  if (nfiles==0)
  {
    printf("repository contains no files.\n");
    exit(0);   
  }

  start_target();
  if (get_remote_file_list(&rfiles, &nrfiles)) 
  {
    printf("can't get remote file index, everything has to be updated\n");
  }

  file_t **extraLO, **extraRE; size_t NextraLO, NextraRE;
  file_t **commLO, **commRE;   size_t NcommLO, NcommRE;
  
  file_set_split(files, nfiles, rfiles, nrfiles,
                 &extraLO, &NextraLO, &commLO, &NcommLO,
                 &extraRE, &NextraRE, &commRE, &NcommRE);

  print_file_list("EXPUNGED FILES", extraRE, NextraRE);
  print_file_list("NEW FILES", extraLO, NextraLO);
  printf("FILES TO BE UPDATED:");
  for(i=0;i<NcommLO;i++)
     if (commLO[i]->mtime > commRE[i]->mtime)
        printf("\n  %s", commLO[i]->name);
  if (i==0) printf(" None");
  printf("\n");
  printf("FILES IN THE FUTURE:");
  for(i=0;i<NcommLO;i++)
     if (commLO[i]->mtime < commRE[i]->mtime)
        printf("\n  %s", commLO[i]->name);
  if (i==0) printf(" None");
  printf("\n");
  end_target();
  exit(0);
}

void do_mark()
{
  file_t **files, **dirs;
  size_t nfiles, ndirs;
  list_all_files(&files, &nfiles, &dirs, &ndirs);
  if (nfiles==0)
  {
    printf("repository contains no files.\n");
    exit(0);   
  }

  start_target();
  printf("marking remote repository as up-to-date\n");
  write_remote_list(files, nfiles);
  end_target();
  exit(0);
}

/* This function updates the given file described in src.
   On success, it stores src in *Rstor and returns zero.
   On failure, it returns 1 and doesn't touch *Rstor.  */
int update_file(file_t **Rstor, file_t *src)
{
  uint8_t *buffer; size_t len;
  printf("updating %s\n", src->name);
  if (make_remote_dirs(src->name)) return 1;
  if (local_read_file(src->name, &buffer, &len))
     goto fail;
  if (write_remote_file(src->name, buffer,len))
     goto fail;
  free(buffer);
  *Rstor= src;
  return 0; 
fail:
  if (buffer) free(buffer);
  return 1;
}

/* This function conditionally updates the remote index.
   It keeps track of bytes/files transferred and updates
   the index if the threshold is met. */
static int update_index(file_t **files, size_t nfiles, size_t fsize,int *Rok)
{
  static size_t x_files, x_bytes;
  size_t V;
  static size_t threshold;
  static int threshold_choice;
  int R;
  x_files++;
  x_bytes+= fsize;

  if (!threshold_choice)
  {
    threshold_choice= 'B';
    switch(target->update.unit)
    {
    case FREQ_KB: threshold= target->update.freq * 1024; break;
    case FREQ_MB: threshold= target->update.freq * 1024 * 1024; break;
    default: threshold= target->update.freq; threshold_choice= 'F';
    }
  }

  if (threshold==0) return 0;

  V= threshold_choice=='F' ? x_files : x_bytes;
  if (V<threshold) return 0;
  *(threshold_choice=='F'? &x_files: &x_bytes) = V % threshold;
  R= write_remote_list(files, nfiles);
  if (!R) *Rok= 1;
  return R;
}

static int interrupted;

void handle_sigint()
{
  printf("interrupting after the current file transfer\n");
  interrupted= 1;
}

static int yesno(const char *msg)
{
  char g[200];
  printf("%s",msg);
  fflush(stdout);
  fgets(g,sizeof(g),stdin);
  if (g[0]!='y' && g[0]!='Y') return 1;
  return 0;
}

static int find_dir(file_t **dirs, int ndirs, char *key)
{
  int i,j,k;
  i= 0; j= ndirs-1;
  while(i<=j)
  {
    k= (i+j)/2;
    switch(sign(strcmp(key,dirs[k]->name)))
    {
    case -1: j= k-1; break;
    case  0: return 1;
    case  1: i= k+1; break;
    }
  }
  return 0;
}

static void remove_remote_extras
      (file_t **ldirs, size_t nldirs, // local repository directories
       file_t **xfiles, size_t nxfiles)
{
  size_t i;
  if (!nxfiles) return ;
  
  for(i=0;i<nxfiles;i++)
  {
    printf("!!EXPUNGED!! %s\n", xfiles[i]->name);
  }
  if (yesno("The above files were recorded in the remote index\n"
            "but no longer exist in the local repository. They\n"
            "will be removed from the remote repository.\n"
            "Do you want to do this? (y/n): ")) return ;

  for(i=0;i<nxfiles;i++)
     remove_remote_file(xfiles[i]->name);

again:
  // Now we transform each file name into a directory name.
  // If the file name doesn't contain a directory component, then
  // we're no longer interested in that, the file itself has already
  // been dealt with.
  for(i=0;i<nxfiles;i++)
  {
    char *P;
    P= strrchr(xfiles[i]->name,'/');
    if (!P)
    {
      xfiles[i]= xfiles[nxfiles-1];
      nxfiles--;
      i--;
      continue;
    }
    *P= 0;
  }
     // all removed files reside at top, which can't be removed.
  if (!nxfiles) return ;
  files_sort(xfiles, nxfiles);
  files_unify(xfiles,&nxfiles);

  for(i=0;i<nxfiles;i++)
  {
    // If the directory exists in the local repository, we aren't
    // going to remove it. Also, we aren't going to remove any of
    // its parents so we can simply take this entry out of the list.
    if (find_dir(ldirs, nldirs, xfiles[i]->name))
    {
      xfiles[i]= xfiles[nxfiles-1];
      nxfiles--;
      i--;
      continue;
    }
    remove_remote_dir(xfiles[i]->name);
  }

  if (!nxfiles) return ;
  // Now, we have considered all entries in the directory list. In order
  // to remove empty parent directories, we take the parent directory
  // of each directory and then unify them. This is the same thing we
  // initially did to the list of file names. So, we can simply jump
  // over there.
  goto again;
}

void do_update()
{
  size_t i;
  file_t **files, **dirs, **rfiles;
  size_t nfiles, ndirs, nrfiles;
  int index_ok;


  list_all_files(&files, &nfiles, &dirs, &ndirs);
  if (nfiles==0)
  {
    printf("repository contains no files.\n");
    exit(0);   
  }

  rfiles= NULL; nrfiles= 0;
  start_target();
  if (get_remote_file_list(&rfiles, &nrfiles)) 
  {
    if (yesno("can't get remote file index, everything has to be updated.\n"
              "do you want to go ahead with this? (y/n): "))
    {
      printf("operation cancelled, exiting.\n");
      end_target();
      exit(0);
    }
  }

  file_t **extraLO, **extraRE; size_t NextraLO, NextraRE;
  file_t **commLO, **commRE;   size_t NcommLO, NcommRE;
  
  file_set_split(files, nfiles, rfiles, nrfiles,
                 &extraLO, &NextraLO, &commLO, &NcommLO,
                 &extraRE, &NextraRE, &commRE, &NcommRE);

  file_t **cfiles; size_t ncfiles;
          // The list will grow as big as the local list.
  cfiles= malloc(sizeof(cfiles[0])*nfiles);
  ncfiles= 0;
  
  index_ok= 1;

  if (NextraRE) index_ok= 0;
  remove_remote_extras(dirs, ndirs, extraRE, NextraRE);

  // Initially, the list will contain only the common 
  // files. Later on, we add the extraLO files. By
  // doing this, we've already removed the entries
  // for expunged files.
  for(i=0;i<NcommLO;i++)
     cfiles[ncfiles++]= commRE[i];

  signal(SIGINT, handle_sigint);

  // Now let's update the common files.
  for(i=0;i<NcommLO;i++)
  {
    uint64_t tL, tR;
    tR= cfiles[i]->mtime;
    tL= commLO[i]->mtime;

    if (interrupted) goto finish_interrupted;

    if (tR==tL) continue;
    if (tL<tR)
    {
       printf("!!FUTURE!! %s\n", cfiles[i]->name);
       continue;
    }

    if (update_file(&cfiles[i], commLO[i])) continue;
    index_ok= 0;
    if (update_index(cfiles, ncfiles, commLO[i]->size,&index_ok))
       goto finish_interrupted; 
  }

  // Time to add the extra files found locally.
  for(i=0;i<NextraLO;i++)
  {
    if (interrupted) goto finish_interrupted;
    if (update_file(&cfiles[ncfiles], extraLO[i]))
       continue;
    index_ok= 0;
    if (update_index(cfiles, ncfiles+1, extraLO[i]->size, &index_ok))
       goto finish_interrupted;
    ncfiles++;
  }

finish_interrupted:
  if (!index_ok)
     write_remote_list(cfiles, ncfiles);
  end_target();
  printf("finished.\n");
  exit(0);
}

int main(int argc, char **argv)
{
  eprintf_attach(writef, stderr, 'c');
  if (argc==2 && !strcmp(argv[1],"help"))
  {
    show_help();
    exit(0);
  }
  read_config();
  parse_args(argc,argv);

  switch(command)
  {
  case 'L': do_list(); break;
  case 'M': do_mark(); break;
  case 'U': do_update(); break;
#if 0
  case 'F': force_compare=1; e_update(); break;
#endif
  }
  return 0;
}

