 /*******************************************************************
  This file handles the target operations. We have two kinds: remote
  and local. Local functions are also used for reading the source
  directory.
 ******************************************************************/
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netdb.h>
#include <libssh2.h>
#include <libssh2_sftp.h>
#include "config.h"
#include "target.h"
#include "eprintf.h"
#include "fnmanip.h"
#include "remodirs.h"


static int block_read
              (void *handle, int (*func)(void *,char *, size_t),
               char **buffer, size_t *len);

 /***********************************************************************
  *                                                                     *
  *                                LOCAL                                *
  *                                                                     *
  * Handles the case where the target is already mounted in some way.   *
  ***********************************************************************/


int local_read_file(char *path, uint8_t **buffer, size_t *len)
{
  int R;
  FILE *f;

  R= 1;
  *buffer= NULL; *len= 0;
  f= fopen(path,"r");
  if (!f)
  {
    eprintf("can't open target %s: %s\n", path, strerror(errno));
    goto E_open;
  }

  R= block_read(f,(void*) local_read, (char**) buffer, len);
  if (R)
    eprintf("error during local-read: %d\n", R);

  fclose(f);

E_open:
  return R;
}

int local_read(void *f,char *Z, size_t len)
{
  int R;
  if (feof(f)) return 0;
  R= fread(Z, 1, len, f);
  if (R<0)
    eprintf("error reading file : %s\n", strerror(errno));
  return R;
}


int local_write_file(char* path, uint8_t *buffer, size_t len)
{
  FILE *f;
  int R;

  R= 1;
  f= fopen(path, "w");
  if (!f) 
  {
    eprintf("can't open file %s for writing: %s\n" , path, strerror(errno));
    goto E_open;
  }

  while(len)
  {
    R= fwrite(buffer, 1, len, f);
    if (R<0) 
    {
      eprintf("error writing file: %s\n", strerror(errno));
      break;
    } 
    if (R==0)
    {
      eprintf("zero writing file?\n");
      break;
    }
    buffer+= R;
    len-= R;
  }
  if (!len) R= 0;

  fclose(f);

E_open:
  return R;
}


 /***********************************************************************
  *                                                                     *
  *                               REMOTE                                *
  *                                                                     *
  ***********************************************************************/

static LIBSSH2_SESSION *SSHsession;
static int SSHfd;
static LIBSSH2_SFTP *SFTPsession;

static int sftp_connect(char *host, char *user);
static int auth_passwd(char *user,char *host);
static void kbd_callback(const char *name, int name_len, 
             const char *instruction, int instruction_len, int num_prompts,
             const LIBSSH2_USERAUTH_KBDINT_PROMPT *prompts,
             LIBSSH2_USERAUTH_KBDINT_RESPONSE *responses,
             void **abstract);
static int auth_kb(char *user);
static int sftp_read_file(char *path, uint8_t **buffer, size_t *len);
static LIBSSH2_SFTP_HANDLE *sftp_open(char *path, int mode);
static int sftp_write_file(char* path, uint8_t *buffer, size_t len);
static void sftp_disconnect();

static int connect_host_tcp4(char *hostname, int port);
static char *sftp_ecode(int code);


static int sftp_connect(char *host, char *user)
{
  int R,i;
  static int libinit_tried;
  char *emsg;
  const char *fingerprint;
  char *userauthlist;

  if (libinit_tried==METHOD_FILE) return 1;

  if (!libinit_tried)
  {
     libinit_tried= 1;
     R= libssh2_init (0);
     if (R)
     {
       libinit_tried++; 
       eprintf("couldn't initialize libssh2: (%d)\n", R);
       return 1;
     }
  }

  SSHfd= connect_host_tcp4(host, 22);
  if (SSHfd<0) goto E_conn;

  SSHsession = libssh2_session_init();
  if(!SSHsession)
  {
     eprintf("error: initialize ssh2 session\n");
     goto E_session1;
  }
  libssh2_session_set_blocking(SSHsession, 1);
  R= libssh2_session_handshake(SSHsession, SSHfd);
  if (R)
  {
    libssh2_session_last_error(SSHsession, &emsg, NULL,0); 
    eprintf("Failure establishing SSH session: (%d) %s\n", R,emsg);
    goto E_handshake;
  }

  fingerprint= libssh2_hostkey_hash(SSHsession, LIBSSH2_HOSTKEY_HASH_SHA1);
  eprintf("Fingerprint: ");
  for(i=0;i<20;i++)
     eprintf("%s%02x", i?":": "", (unsigned char)fingerprint[i]);
  eprintf("\n");

  userauthlist = libssh2_userauth_list(SSHsession, user, strlen(user));
  eprintf("Authentication methods: %s\n", userauthlist);

  if (strstr(userauthlist, "password") && auth_passwd(user,host))
    goto auth_OK;
  if (strstr(userauthlist, "keyboard-interactive") && auth_kb(user))
    goto auth_OK;

  eprintf("No usable authentication methods.\n");
  goto E_auth;

auth_OK:
  SFTPsession = libssh2_sftp_init(SSHsession);

  if (!SFTPsession)
  {
    eprintf("Unable to init SFTP session\n");
    goto E_auth;
  }

  eprintf("sftp connection OK.\n");
  return 0;

E_auth:
  libssh2_session_disconnect(SSHsession, "client got confused");
E_handshake:
  libssh2_session_free(SSHsession);
E_session1:
  close(SSHfd);
E_conn:
  libssh2_exit();
  libinit_tried= 2;
  return 1;
}


static int auth_passwd(char *user,char *host)
{
   char *Z;
   char g[200];
   printf("password for %s@%s:",user,host);
   fflush(stdout);
   fgets(g, 200, stdin);
   Z= strchr(g,'\r'); if (Z) Z[0]= 0;
   Z= strchr(g,'\n'); if (Z) Z[0]= 0;

   if (libssh2_userauth_password(SSHsession, user, g))
   {
     eprintf("password authentication failed.\n");
     return 0;
   }
   return 1;
}


static void kbd_callback(const char *name, int name_len, 
             const char *instruction, int instruction_len, int num_prompts,
             const LIBSSH2_USERAUTH_KBDINT_PROMPT *prompts,
             LIBSSH2_USERAUTH_KBDINT_RESPONSE *responses,
             void **abstract)
{
    int i;
    size_t n;
    char buf[1024];
    (void)abstract;

    fprintf(stderr, "Performing keyboard-interactive authentication.\n");

    fprintf(stderr, "Authentication name: '");
    fwrite(name, 1, name_len, stderr);
    fprintf(stderr, "'\n");

    fprintf(stderr, "Authentication instruction: '");
    fwrite(instruction, 1, instruction_len, stderr);
    fprintf(stderr, "'\n");

    fprintf(stderr, "Number of prompts: %d\n\n", num_prompts);

    for (i = 0; i < num_prompts; i++) {
        fprintf(stderr, "Prompt %d from server: '", i);
        fwrite(prompts[i].text, 1, prompts[i].length, stderr);
        fprintf(stderr, "'\n");

        fprintf(stderr, "Please type response: ");
        fgets(buf, sizeof(buf), stdin);
        n = strlen(buf);
        while (n > 0 && strchr("\r\n", buf[n - 1]))
          n--;
        buf[n] = 0;

        responses[i].text = strdup(buf);
        responses[i].length = n;

        fprintf(stderr, "Response %d from user is '", i);
        fwrite(responses[i].text, 1, responses[i].length, stderr);
        fprintf(stderr, "'\n\n");
    }

    fprintf(stderr,
        "Done. Sending keyboard-interactive responses to server now.\n");
}


static int auth_kb(char *user)
{
  if (libssh2_userauth_keyboard_interactive
        (SSHsession, user, &kbd_callback))
  {
    eprintf("interactive authentication failed.\n");
    return 0;
  }
  return 1;
}

static int sftp_read_file(char *path, uint8_t **buffer, size_t *len)
{
  LIBSSH2_SFTP_HANDLE *handle;
  int R;

  *buffer= NULL; *len= 0;

  handle= sftp_open(path, 'r');
  R= 1;
  if (!handle) goto E_open;

  R= block_read(handle,(void*) libssh2_sftp_read, (char**) buffer, len);
  if (R)
  {
    eprintf("error during sftp-read: %d\n", R);
  }

  libssh2_sftp_close(handle);

E_open:
  return R;
}


static LIBSSH2_SFTP_HANDLE *sftp_open(char *path, int mode)
{
  LIBSSH2_SFTP_HANDLE *handle;
  int smode;

  if (mode=='r') smode= LIBSSH2_FXF_READ;
  else           smode= LIBSSH2_FXF_WRITE | LIBSSH2_FXF_CREAT |
                        LIBSSH2_FXF_TRUNC;
  
  handle = libssh2_sftp_open_ex(SFTPsession, path, strlen(path),
                     smode, 0644, LIBSSH2_SFTP_OPENFILE);
  if (!handle)
  {
    eprintf("can't open %s over SFTP: %d\n",path,
              sftp_ecode(libssh2_sftp_last_error(SFTPsession)));
    return NULL;
  }
  return handle;
}


static int sftp_write_file(char* path, uint8_t *buffer, size_t len)
{
  LIBSSH2_SFTP_HANDLE *handle;
  int R;
  handle= sftp_open(path, 'w');
  R= 1;
  if (!handle) goto E_open;

  while(len)
  {
    R= libssh2_sftp_write(handle, (const char*) buffer, len);
    if (R<0) 
    {
      eprintf("error writing file: %d\n", R);
      break;
    } 
    if (R==0)
    {
      eprintf("zero writing file?\n");
      continue;
    }
    buffer+= R;
    len-= R;
  }
  if (!len) R= 0;

  libssh2_sftp_close(handle);

E_open:
  return R;
}

static void sftp_disconnect()
{
   libssh2_sftp_shutdown(SFTPsession);
   libssh2_session_disconnect(SSHsession,
                           "Normal Shutdown, Thank you for playing");
   libssh2_session_free(SSHsession);
   close(SSHfd);
}

static int connect_host_tcp4(char *hostname, int port)
{
  struct addrinfo hints;
  struct addrinfo *res, *ai;
  char portbuf[20];
  int R,fd;

  sprintf(portbuf, "%d", port);

  memset(&hints, 0, sizeof(struct addrinfo));
  hints.ai_family = AF_INET;
  hints.ai_socktype = SOCK_STREAM;
  hints.ai_flags = AI_NUMERICSERV;
  hints.ai_protocol = 0;

  R= getaddrinfo(hostname, portbuf, &hints, &res);
  if (R)
  {
    eprintf("can't resolve %s: %s\n", hostname, gai_strerror(R));
    return -1;
  }

  for(ai=res;ai;ai=ai->ai_next)
  {
    fd= socket(ai->ai_family, ai->ai_socktype, ai->ai_protocol);
    if (fd<0)
    {
      eprintf("can't create socket: %s\n", strerror(errno));
      continue;
    }
    eprintf("Trying to connect %s : ", hostname);
    if (connect(fd, ai->ai_addr, ai->ai_addrlen))
    {
       eprintf(" %s\n", strerror(errno));
       close(fd);
    }
    else 
    {
       eprintf(" OK\n");
       freeaddrinfo(res);
       return fd;
    }
  }
  freeaddrinfo(res);
  eprintf("couldn't connect with any address\n");
  return -1;
}

void start_target()
{
  if (target->method==METHOD_FILE) return ;
  if (sftp_connect(target->host, target->user)) exit(1);
}

void end_target()
{
  if (target->method==METHOD_FILE) return ;
  sftp_disconnect();
}

int read_remote_file(char *name, uint8_t **buffer, size_t *len)
{
   char *fn; int R;
   fn= fnmani_child(target->dir, name);
   if (target->method==METHOD_FILE) R= local_read_file(fn,buffer,len);
   else R= sftp_read_file(fn,buffer,len);
   free(fn);
   return R;
}

int write_remote_file(char *name, uint8_t *buffer, size_t len)
{
   char *fn; int R;
   fn= fnmani_child(target->dir, name);
   if (target->method==METHOD_FILE) R= local_write_file(fn,buffer,len);
   else R= sftp_write_file(fn,buffer,len);
   free(fn);
   return R;
}



static char *sftp_ecode(int code)
{
switch(code) {
case LIBSSH2_FX_OK                    : return "ok";
case LIBSSH2_FX_EOF                   : return "eof";
case LIBSSH2_FX_NO_SUCH_FILE          : return "no such file";
case LIBSSH2_FX_PERMISSION_DENIED     : return "permission denied";
case LIBSSH2_FX_FAILURE               : return "failure";

case LIBSSH2_FX_BAD_MESSAGE           : return "bad message";
case LIBSSH2_FX_NO_CONNECTION         : return "no connection";
case LIBSSH2_FX_CONNECTION_LOST       : return "connection lost";
case LIBSSH2_FX_OP_UNSUPPORTED        : return "op unsupported";
case LIBSSH2_FX_INVALID_HANDLE        : return "invalid handle";

case LIBSSH2_FX_NO_SUCH_PATH          : return "no such path";
case LIBSSH2_FX_FILE_ALREADY_EXISTS   : return "file already exists";
case LIBSSH2_FX_WRITE_PROTECT         : return "write protect";
case LIBSSH2_FX_NO_MEDIA              : return "no media";
case LIBSSH2_FX_NO_SPACE_ON_FILESYSTEM: return "no space on filesystem";

case LIBSSH2_FX_QUOTA_EXCEEDED        : return "quota exceeded";
case LIBSSH2_FX_UNKNOWN_PRINCIPAL     : return "unknown principal";
case LIBSSH2_FX_LOCK_CONFLICT         : return "lock conflict";
case LIBSSH2_FX_DIR_NOT_EMPTY         : return "dir not empty";
case LIBSSH2_FX_NOT_A_DIRECTORY       : return "not a directory";

case LIBSSH2_FX_INVALID_FILENAME      : return "invalid filename";
case LIBSSH2_FX_LINK_LOOP             : return "link loop";
}
return "unknown error";
}

 /***********************************************************************

                               BLOCK READ                              

  The following function reads a given 'stream' without calling stat() on
  a file. The given function should return a negative value for an error,
  zero for end of file, and a positive value for a successful read.

  The return value of the block_read function is zero for success and
  non-zero for an error, which will be returned by 'func'. The allocated
  buffer has an extra space in which a null character will be stored.

  The contents read will be retained even if an error is detected along
  the way.
  ***********************************************************************/

static int block_read_real 
              (void *handle, int (*func)(void *,char *, size_t),
               char **buffer, size_t *len);

static int block_read
              (void *handle, int (*func)(void *,char *, size_t),
               char **buffer, size_t *len)
{
  size_t L;
  int R;
  L= 0;
  *buffer= NULL;
  R= block_read_real(handle, func, buffer, &L);
  (*buffer)[L]= 0;
  if (len) *len= L;
  return R;
}

static int block_read_real 
              (void *handle, int (*func)(void *,char *, size_t),
               char **buffer, size_t *len)
{
    char *xb;
    static const int rsize=2000;
    int R,B;
    size_t pos= *len;

    B= 0;
    xb= malloc(rsize);
    R= (*func)(handle, xb, rsize);
    if (R<=0)    goto eof_or_error;
    
    B= R;
    *len+= B;
    R= block_read_real(handle, func, buffer, len);

    memcpy(*buffer + pos, xb, B);
    free(xb);
    return R;
eof_or_error:
    free(xb);
    *buffer= malloc(*len+1);
    return 0;
}

/* This function first stat()s the corresponding name and then
   tries to create a directory if that fails. */
int target_mkdir(char *name)
{
   struct stat sb;
   LIBSSH2_SFTP_ATTRIBUTES sfb;

   name= fnmani_child(target->dir, name);

   if (target->method==METHOD_FILE)
   {
     if (stat(name, &sb)) 
     {
       if (mkdir(name,0755)) 
       {
         eprintf("local mkdir(%s) failed: [%d] %s\n", 
                    name, errno,strerror(errno));
         goto fail;
       }
       goto success;
     }
     if (S_ISDIR(sb.st_mode)) goto success;
     eprintf("local target %s exists, but isn't a directory\n", name);
     goto fail;
   }

   if (libssh2_sftp_stat(SFTPsession, name, &sfb))
   {
     if (libssh2_sftp_mkdir(SFTPsession, name, 0755))
     {
       eprintf("can't make remote directory %s\n", name);
       goto fail;
     }
     goto success;
   }
   else
   {
     if (LIBSSH2_SFTP_S_ISDIR(sfb.permissions)) goto success;
     eprintf("remote target %s exists, but isn't a directory\n",name);
     goto fail;
   }

fail:
   free(name);
   return 1;
success:
   free(name);
   return 0;
}


int remove_remote_file(char *name)
{
  char *fn;
  printf("removing target file %s\n", name);
  fn= fnmani_child(target->dir, name);
  if (target->method==METHOD_FILE)
  {
    if (unlink(fn))
    {
      eprintf("can't remove file %s: [%d] %s\n", fn, errno, strerror(errno));
      goto fail;
    }
    goto success;
  }

  if (libssh2_sftp_unlink(SFTPsession, fn))
  {
    eprintf("[sftp] can't remove file %s\n", fn);
    goto fail;
  }

success:
  free(fn);
  return 0;
fail:
  free(fn);
  return 1;
}

int remove_remote_dir(char *name)
{
  char *fn;
  printf("removing target directory %s\n", name);
  fn= fnmani_child(target->dir, name);
  if (target->method==METHOD_FILE)
  {
    if (rmdir(fn))
    {
      eprintf("can't remove directory %s: [%d] %s\n",
               fn, errno, strerror(errno));
      goto fail;
    }
    goto success;
  }

  if (libssh2_sftp_rmdir(SFTPsession, fn))
  {
    eprintf("[sftp] can't remove directory %s\n", fn);
    goto fail;
  }

success:
  free(fn);
  return 0;
fail:
  free(fn);
  return 1;
}
